# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" UI class and functions """


import panel as pn
import matplotlib.pyplot as plt
import datetime

from med_diagnostics import data
from IPython.display import display

class UserInterface():
    
    """
    Primary class for user interface (UI) components and deployment
    """
    
    def __init__(self):
        
        """
        Initialise a UserInterface instance.
        """

        # Import panel extensions
        pn.extension()
        
        # Build and style panel widgets
        self.last_data_load_textbox = pn.widgets.StaticText(styles={'background': 'orange', 'font-size': '18px', 'color': 'black', 'padding': '5px'}, margin=(10, 0, 10, 0))
        self.status_textbox = pn.widgets.StaticText(styles={'background': 'lightblue', 'font-size': '18px', 'color': 'black', 'padding': '5px'}, margin=(10, 0, 10, 0))
        self.ref_status_textbox = pn.widgets.StaticText(styles={'background': 'lightblue', 'font-size': '18px', 'color': 'black', 'padding': '5px'}, margin=(10, 0, 10, 0))

        self.ref_model_metadata = pn.widgets.StaticText(styles={'color': 'white'})
        
        self.keys_dropdown = pn.widgets.Select()
        self.keys_button = pn.widgets.Button(styles={}, margin=(23, 0, 0, 0))
        self.plot_variable_dropdown = pn.widgets.Select()
        
        self.ref_keys_dropdown = pn.widgets.Select()
        self.ref_keys_button = pn.widgets.Button(styles={}, margin=(23, 0, 0, 0))
        self.ref_plot_variable_dropdown = pn.widgets.Select()
        
        self.ref_data_keys_dropdown = pn.widgets.Select()
        self.ref_data_keys_button = pn.widgets.Button(styles={}, margin=(23, 0, 0, 0))
        self.ref_data_plot_variable_dropdown = pn.widgets.Select()

        self.clear_ref_model_data_button = pn.widgets.Button(styles={}, margin=(23, 0, 0, 10))
        self.ref_model_info_button = pn.widgets.Button(styles={}, margin=(23, 10, 0, 0))
        
        self.figure_exists = False
        self.ref_figure_exists = False
        
        
        # Initialise button listener functions
        @pn.depends(self.keys_dropdown.param.value)
        def _keys_button_click(event):
            
            self._keys_dropdown_click()
            
        @pn.depends(self.ref_keys_dropdown.param.value)
        def _ref_keys_button_click(event):
            
            self._ref_keys_dropdown_click()
            
        @pn.depends(self.ref_data_keys_dropdown.param.value)
        def _ref_data_keys_button_click(event):
            
            self._ref_dataset_dropdown_click()

        @pn.depends(self.ref_keys_dropdown.param.value)
        def _ref_model_info_button_click(event):
            
            self._ref_model_info_click()

        def _ref_clear_data_button_click(event):
            
            self._ref_clear_data_click()
            
        self.keys_button.on_click(_keys_button_click)
        self.ref_keys_button.on_click(_ref_keys_button_click)
        self.ref_data_keys_button.on_click(_ref_data_keys_button_click)
        self.clear_ref_model_data_button.on_click(_ref_clear_data_button_click)
        self.ref_model_info_button.on_click(_ref_model_info_button_click)

            
            
    def _display_status_text(self):
        
        """
        Create widget_container then add status_textbox and last_data_load_textbox widgets. Private.
        """
        
        # Create panel column
        self.widget_container = pn.Column()
        
        # Append widget_container with textbox widgets
        self.widget_container.append(self.last_data_load_textbox)
        self.widget_container.append(self.status_textbox)
        
        # Display widget_container in notebook
        display(self.widget_container)
        print()
        
        
    def _update_status_text(self, text):
        
        """
        Update text displayed in status_textbox widget. Private.
        
        Parameters
        ----------
        text : str
            Text to be displayed in status_textbox
        """
        
        # Update status_textbox with text
        self.status_textbox.value = str(text)


    def _update_ref_status_text(self, text):
        
        """
        Update text displayed in ref_status_textbox widget. Private.
        
        Parameters
        ----------
        text : str
            Text to be displayed in ref_status_textbox
        """

        # Update ref_status_textbox with text
        self.ref_status_textbox.value = str(text)
        
        
    def _update_last_data_load_text(self, text):
        
        """
        Update text displayed in last_data_load_textbox widget. Private.
        
        Parameters
        ----------
        text : str
            Text to be displayed in last_data_load_textbox
        """
        
        # Update last_data_load_textbox with text
        self.last_data_load_textbox.value = str(text)
        
        
    def _get_current_time(self):
        
        """
        Get current time. Private.
        
        Returns
        ----------
        str
            Current time in "%Y-%m-%d %H:%M:%S" format.
        """
        
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        
        
    def _display_dataset_selection_ui(self, model_cat, access_nri_cat):
        
        """
        Label, populate and append dataset selection-related widgets to widget_container. Private.
        
        Parameters
        ----------
        model_cat : Intake-ESM datastore object
            Intake catalog of user model data.
        access_nri_cat : Intake-ESM datastore object
            Intake catalog of ACCESS model data.
        """
        
        # Assign argument to class-accessible variables
        self.model_cat = model_cat
        self.access_nri_cat = access_nri_cat
        
        # Add horizontal line divider to widget_container
        self.widget_container.append(pn.layout.Divider(styles={'color': 'white'}))
        
        # Populate user model widgets
        self.keys_dropdown.name = '1. Please select a dataset to monitor:'
        self.keys_dropdown.options = self.model_cat.keys()
        self.keys_button.name = 'Load dataset'
        self.keys_button.button_type = 'primary'
        
        # Add user model widgets to keys_selection_row
        self.keys_selection_row = pn.Row()
        self.keys_selection_row.append(self.keys_dropdown)
        self.keys_selection_row.append(self.keys_button)
        
        # Add keys_selection_row to widget_container
        self.widget_container.append(self.keys_selection_row)
        
        # Add horizontal line divider to widget_container
        self.widget_container.append(pn.layout.Divider(styles={'color': 'white'}))

        #print(self.widget_container)
        
        
    def _display_reference_model_selection_ui(self):
        
        """
        Label, populate and append ACCESS reference model selection-related widgets to widget_container. Private.
        
        """

        # Add refrence data status text box
        self.widget_container.append(self.ref_status_textbox)
        
        # Populate reference/comparison model widgets
        self.ref_keys_dropdown.name = '2. Select reference model (optional):'
        self.ref_keys_dropdown.options = self.access_nri_cat.keys()
        self.ref_keys_button.name = 'Load reference model'
        self.ref_keys_button.button_type = 'primary'
        
        self.clear_ref_model_data_button.name = 'Clear reference model'
        self.clear_ref_model_data_button.button_type = 'primary'

        self.ref_model_info_button.name = 'Reference model information'
        self.ref_model_info_button.button_type = 'primary'
        
        # Add reference/comparison widgets to ref_keys_selection_row
        self.ref_keys_selection_row = pn.Row()
        self.ref_keys_selection_row.append(self.ref_keys_dropdown)
        self.ref_keys_selection_row.append(self.ref_model_info_button)
        self.ref_keys_selection_row.append(self.ref_keys_button)
        self.ref_keys_selection_row.append(self.clear_ref_model_data_button)
        
        # Add ref_keys_selection_row to widget_container
        self.widget_container.append(self.ref_keys_selection_row)
        self.widget_container.append(self.ref_model_metadata)
        
        # Add horizontal line divider to widget_container
        self.widget_container.append(pn.layout.Divider(styles={'color': 'white'}))
        
        
    def _display_reference_dataset_selection_ui(self):
        
        """
        Label, populate and append ACCESS reference dataset selection-related widgets to widget_container. Private.
        
        """
        
        # Populate reference/comparison dataset widgets
        self.ref_data_keys_dropdown.name = '2.1. Select reference dataset (optional):'
        self.ref_data_keys_dropdown.options = self.ref_model_cat.keys()
        self.ref_data_keys_button.name = 'Load reference dataset'
        self.ref_data_keys_button.button_type = 'primary'
        
        # Add reference/comparison widgets to ref_data_keys_selection_row
        self.ref_data_keys_selection_row = pn.Row()
        self.ref_data_keys_selection_row.append(self.ref_data_keys_dropdown)
        self.ref_data_keys_selection_row.append(self.ref_data_keys_button)
        
        # Add ref_data_keys_selection_row to widget_container
        self.widget_container.append(self.ref_data_keys_selection_row)
        
        # Add horizontal line divider to widget_container
        self.widget_container.append(pn.layout.Divider(styles={'color': 'white'}))
        
        
        
    def _keys_dropdown_click(self):
        
        """
        Loads selected model dataset from keys_dropdown and creates new interactive plot. Private.
        """
        
        # Update text box
        self._update_status_text('User model status >> Loading data.')

        # Load selected dataset
        self.dataset = data._build_data_object(self.model_cat, self.keys_dropdown.value)
        
        # Update text box
        self._update_status_text('User model status >> Data successfully loaded.')
        
        # Check if plot already exists
        if self.figure_exists == False:

            # Create new plot
            self._display_dataset_plot_ui()
            
            # Add ui for loading optional reference dataset
            self._display_reference_model_selection_ui()
            
        elif self.figure_exists == True:

            # Update existing plot
            self._update_dataset_plot_ui()
            
            
    def _ref_keys_dropdown_click(self):
        
        """
        Loads selected reference model from ref_keys_dropdown and display reference model dataset selection. Private.
        """

        # Update text box
        self._update_ref_status_text('Reference model status >> Loading data.')
        
        # Extract selected model catalog
        self.ref_model_cat = self.access_nri_cat.search(name=self.ref_keys_dropdown.value).to_source()

        # Update text box
        self._update_ref_status_text('Reference model status >> Data successfuly loaded.')

        # Create new plot
        self._display_reference_dataset_selection_ui()
                    
        
        
    def _ref_dataset_dropdown_click(self):
        
        """
        Loads selected reference model dataset from ref_data_keys_dropdown and creates new interactive plot. Private.
        """
        
        # Load selected access_nri catalog dataset
        self.ref_dataset = data._build_data_object(self.ref_model_cat, self.ref_data_keys_dropdown.value)

        # Create new plot
        self._display_ref_dataset_plot_ui()
            


    def _ref_clear_data_click(self):
        """
        Clears and 'unloads' selected reference model dataset from widget container. Private.
        """

        # Update text box
        self._update_ref_status_text('Reference model status >> Data removed.')

        # Remove reference model widgets from container one at a time
        self.widget_container.pop(-1)
        self.widget_container.pop(-1)
        self.widget_container.pop(-1)
        self.widget_container.pop(-1)

        # Remove reference model ref_model_cat
        del self.self.ref_model_cat
        
        # Remove reference dataset ref_dataset
        del self.ref_dataset


    def _ref_model_info_click(self):

        # Update text box
        self._update_ref_status_text('Reference model status >> Retrieving model metadata.')

        # self.ref_model_metadata.value = self.access_nri_cat[self.ref_keys_dropdown.value].metadata['description']

        self.ref_model_metadata.value = '<b>Model information:</b><br>' + \
                                        '<b>Model:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['model']) + '<br>' + \
                                        '<b>Short description:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['description']) + '<br>' + \
                                        '<b>Nominal resolution:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['nominal_resolution']) + '<br>' + \
                                        '<b>Parent experiment:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['parent_experiment']) + '<br>' + \
                                        '<b>Long description:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['long_description']) + '<br>' + \
                                        '<b>Contact:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['contact']) + '<br>' + \
                                        '<b>Email:   </b>' + str(self.access_nri_cat[self.ref_keys_dropdown.value].metadata['email']) + '<br>'

        # Update text box
        self._update_ref_status_text('')
        
    def _display_dataset_plot_ui(self):
        
        """
        Create interactive panel plot for user model dataset and add to widget_container. Private.
        """
        
        self.plot_variable_dropdown.name = 'Available variables'
        self.plot_variable_dropdown.options = list(self.dataset.keys())
        
        self.interactive_plot = pn.bind(self._plot_dataset, self.plot_variable_dropdown)
        
        self.widget_container.append(self.plot_variable_dropdown)
        self.widget_container.append(self.interactive_plot)
        
        
    def _display_ref_dataset_plot_ui(self):
        
        """
        Create interactive panel plot for reference model dataset and add to widget_container. Private.
        """
        
        self.ref_plot_variable_dropdown.name = 'Available reference variables'
        self.ref_plot_variable_dropdown.options = list(self.ref_dataset.keys())
        
        self.ref_interactive_plot = pn.bind(self._plot_dataset, self.plot_variable_dropdown, self.ref_plot_variable_dropdown)
        
        self.widget_container.append(self.ref_plot_variable_dropdown)
        self.widget_container.append(self.ref_interactive_plot)
        
        
    def _update_dataset_plot_ui(self):
        
        """
        Update exisiting user model dataset plot if new data are selected. Private.
        """
        
        self.plot_variable_dropdown.options = list(self.dataset.keys())
        plt.close(self.fig)
        
        
    def _update_ref_dataset_keys_plot_ui(self):
        
        """
        Update exisiting reference model dataset keys if new data are selected. Private.
        """
        
        self.ref_data_keys_dropdown.options = list(self.ref_model_cat.keys())
        plt.close(self.ref_fig)
        
        self._update_ref_dataset_plot_ui()
       
        
    def _update_ref_dataset_plot_ui(self):
        
        """
        Update exisiting reference model dataset plot if new data are selected. Private.
        """
        
        self.ref_plot_variable_dropdown.options = list(self.ref_dataset.keys())
        self.ref_plot_variable_dropdown.value = list(self.ref_dataset.keys())[0]
        plt.close(self.ref_fig)
        
        
    def _plot_dataset(self, variable, ref_variable=None):
        
        """
        Plot 2D time-series from model data. Private.
        
        Parameters
        ----------
        variable : str
            Model data variable as selected from panel dropdown.
        ref_variable : str, default None
            Reference model data variable as selected from panel dropdown.
            
        Returns
        ----------
        fig : matplotlib.pyplot.figure()
        ref_fig : matplotlib.pyplot.figure()
        """
        
        if ref_variable == None:
            
            # Plot primary (user) model data
            self.fig = plt.figure(figsize=[8,4])
            
            self.figure_exists = True
            
            self.dataset[variable].plot()
            
            plt.title(self.dataset[variable].attrs['long_name'], fontsize=14)

            plt.grid()
            plt.legend()
            plt.tight_layout()

            plt.close(self.fig)

            return self.fig
            
        else:
            
            # Plot reference model
            self.ref_fig = plt.figure(figsize=[8,4])
            
            self.ref_figure_exists = True
            
            # Plot all model variants if multiple exist
            if "member" in self.ref_dataset.dims:

                for mem in self.ref_dataset.member.values:

                    self.ref_dataset[ref_variable].sel(member=mem).plot(label=mem)
            
            plt.title(self.ref_dataset[ref_variable].attrs['long_name'], fontsize=14)

            plt.grid()
            plt.legend()
            plt.tight_layout()

            plt.close(self.ref_fig)

            return self.ref_fig

        
        



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Session class functions for med-diagnostics live diagnostics """


import os
import intake
import panel as pn

from apscheduler.schedulers.background import BackgroundScheduler
from med_diagnostics import data, ui
from distributed import Client


class CreateModelDiagnosticsSession():
    
    """
    Primary class for starting a model diagnostics session
    """
    
    def __init__(self, model_type, model_path, period=None, timezone=None):
        
        """
        Initialise a CreateLiveSession instance to start a model diagnostics session.

        Parameters
        ----------
        model_type : str
            Type of ACCESS model in capitals (e.g. CM2, OM2).
        model_path : str
            Path to model output directory/files on Gadi.
        period : int/float/str, optional, default None
            Period in minutes for background scheduler to monitor model_path for new data. If period=None, defaults to 60.
        timezone : str, optional, default 'Australia/Canberra'
            Timezone required for scheduler in tinfo 'Region/Location' format. 
            
        """
        
        # Set local variables
        self.model_type = str(model_type).lower()
        # self.model_realm = str(model_realm)
        self.model_path = str(model_path)
        self.model_data = []
        
        self.period = float(period) if period != None else 60.0
        self.timezone = str(timezone) if timezone != None else 'Australia/Canberra'
        
        self.data_update = False
        
        # Start dask client
        self.client = Client(threads_per_worker=1)
        
        print()
        print('----------------------- Live diagnostics session started -----------------------')
        print()
        print('Model type:', str(model_type))
        print('Model data path:', self.model_path)
        print('Model data update period:', self.period, 'mins')
        print()
        print('Started dask client:', self.client.dashboard_link)
        print()
        print('--------------------------------------------------------------------------------')
        print()
        
        # Start UserUI instance and display initial status text
        self.ui = ui.UserInterface()
        self.ui._display_status_text()
        
        # Start data scheduler
        self._start_scheduler()
        
        # Get initial model data
        self._get_data()

        
    def _start_scheduler(self):
        
        """
        Start background scheduler to trigger model data retrieval function get_data() at nominated period. Private.
        """

        self.scheduler = BackgroundScheduler()
        self.scheduler.add_job(self._get_data, 'interval', minutes=self.period, timezone=self.timezone)

        self.scheduler.start()
        
        
    def end_session(self):

        """
        Stop background scheduler and close dask client to end current CreateModelDiagnosticsSession instance.
        """

        self.scheduler.shutdown()
        self.client.close()

        self.ui.widget_container.clear()
        
        print('------------------------ Live diagnostics session ended ------------------------')
        
        
    def _get_data(self):
        
        """
        Check nominated model data path for new data. Private.
        """
        
        self.ui._update_status_text('User model status >> Building initial model data catalog. This can take a few minutes.')
        
        # Check for new data
        new_model_data = data._check_for_new_data(self.model_path, self.model_data, self.model_type)
        
        # Update status text
        self.ui._update_status_text('User model status >> Model data catalog built.')
        self.ui._update_last_data_load_text('Last model data catalog build >> ' + self.ui._get_current_time())
        
        if new_model_data == None:
            
            # Do nothing
            pass
            
        else:
            
            # Data loading procedure for initial step
            if self.data_update == False:
            
                # Update self.model_data with new data
                self.model_data = new_model_data

                # Load new catalog
                self.model_cat = data._load_new_catalog()
                
                # Load access_nri catalog for model comparison filtered by model type
                self.access_nri_cat = data._load_access_nri_catalog(self.model_type)

                # Generate UI
                self.ui._display_dataset_selection_ui(self.model_cat, self.access_nri_cat)
            
                self.data_update == True
                
                
            # Data loading procedure for update step
            elif self.data_update == True:
                
                self.ui._update_status_text('User model status >> New data found. Please restart session to view new data.')
                
        
    def return_model_data_catalog(self):
        
        """
        Convenience function to return currently loaded model data catalog.
        
        Returns
        ----------
        Intake-ESM datastore object
            Intake catalog of user model data.
        """
        
        return self.model_cat
    
    
    

        



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Data IO functions """


import os
import time
import intake

from access_nri_intake.source.builders import AccessCm2Builder # for use with access-med-env


        
def _check_for_new_data(model_path, model_data, model_type):
    
    """
    Check monitored directory for new data and build new catalog if found.
    
    Parameters
    ----------
    model_path : str
        Path to model output directory/files on Gadi.
    model_data : list
        List of datafiles present in nominated source directory.
        
    Returns
    ----------
    new_model_data : intake ESM datastore
        Returns intake ESM datastore of user model data if data found.
    """
    
    # Return contents of nominated model data directory
    current_model_data = [f for f in os.listdir(model_path) if os.path.isfile(os.path.join(model_path, f))]
    
    # Check if any files have changed
    if current_model_data != model_data:
        
        # Set new_model_data to current_model_data
        new_model_data = current_model_data
        
        # Build ESM catalog to include new data
        _build_new_catalog(model_path, model_type)
        
        return new_model_data
        
    else:
        
        return None
    

        
def _build_new_catalog(model_path, model_type):
    
    """
    Build new intake ESM datastore from user model data found in model_path. 
    Saves new catalog file to user home. Private.
    
    Parameters
    ----------
    model_path : str
        Path to model output directory/files on Gadi.
    model_type : str
        Type of ACCESS model (e.g. CM2, OM2).
    """
    
    # Use CM2 builder
    if model_type == 'cm2':
    
        builder = AccessCm2Builder(path=model_path, ensemble=False).build()
        builder.save(name="live_diagnostics_tmp_catalog", description="Temporary catalog for live diagnostics", directory=os.getcwd())
    
    
def _load_new_catalog():
            
    """
    Load saved ESM datastore catalog file from user home path. Private.
    
    Returns
    ----------
    model_cat : Intake-ESM datastore object
        Intake catalog of user model data.
    """

    model_cat = intake.open_esm_datastore(os.path.join(os.getcwd(), "live_diagnostics_tmp_catalog.json"), columns_with_iterables=["variable"])
    
    return model_cat


def _start_dask_cluster():
    
    """
    Starts local dask cluster for data retrieval. Private.
    
    Returns
    ----------
    client.dashboard_link : str
        Dask client 'dashboard url' for monitoring.
    """
    
    from distributed import Client

    client = Client(threads_per_worker=1)
    
    return(client.dashboard_link)


def _build_data_object(model_cat, key):
    
    """
    Covert model_cat ESM datastore to xarray object. Private.
    
    Parameters
    ----------
    model_cat : Intake-ESM datastore object
        Intake catalog of user model data.
    key : str
        model_cat dictionary key - in this case the selected dropdown value.
        
    Returns
    ----------
    dataset : xarray object
        Dask xarray object.
    """
    
    # Load dataset using dask to xarray object
    dataset = model_cat[key](xarray_open_kwargs=dict(use_cftime=True)).to_dask()
    
    return dataset


def _load_access_nri_catalog(model_type, filter=True):
    
    """
    Load ACCESS-NRI data catalog. Private.
    
    Parameters
    ----------
    model_type : str
        Type of ACCESS model (e.g. CM2, OM2).
        
    Returns
    ----------
    ACCESS-NRI intake catalog
    """
    
    catalog = intake.cat.access_nri
    
    if filter == False:
        
        return catalog
    
    else:
        
        # Filter catalog by model type
        return catalog.search(model='.*' + model_type.upper() + '.*')
